#include <iostream>
#include <iomanip>
#include <algorithm>
#include <list>
#include <unordered_set>
#include "MeshLoader.h"

std::vector<Node>& MeshLoader::Get_Nodes() {
	return nodes;
}

std::vector<FiniteElement>& MeshLoader::Get_Elements() {
	return elements;
}

std::vector<BoundaryFiniteElement>& MeshLoader::Get_Surfaces() {
	return boundary;
}

void MeshLoader::Print_Data() {
	std::cout << nodes << std::endl
		<< elements << std::endl
		<< boundary << std::endl;
}


//!!! Контейнеры должны возвращать ID, а не сами узлы и КЭ +


std::vector<FiniteElement> MeshLoader::Get_Elements_by_ID(int id1, int id2, int id3) {
	std::vector<FiniteElement> answer;
	auto current_iter = elements.begin();
	while (current_iter != elements.end()) {
		current_iter = std::find_if(current_iter, elements.end(), [id1, id2, id3](const FiniteElement& current_element) {
			return std::find(current_element.nodes_id.begin(), current_element.nodes_id.end(), id1) != current_element.nodes_id.end()
				&& std::find(current_element.nodes_id.begin(), current_element.nodes_id.end(), id2) != current_element.nodes_id.end()
				&& std::find(current_element.nodes_id.begin(), current_element.nodes_id.end(), id3) != current_element.nodes_id.end(); });
		if (current_iter != elements.end()) {
			answer.push_back(*current_iter);
			++current_iter;
		}
	}
	return answer;
}

std::vector<FiniteElement> MeshLoader::Get_Elements_by_rib(int id1, int id2) {
	std::vector<FiniteElement> answer;

	auto current_iter = elements.begin();
	while (current_iter != elements.end()) {
		current_iter = std::find_if(current_iter, elements.end(), [id1, id2](const FiniteElement& current_element) {
			return std::find(current_element.nodes_id.begin(), current_element.nodes_id.end(), id1) != current_element.nodes_id.end()
				&& std::find(current_element.nodes_id.begin(), current_element.nodes_id.end(), id2) != current_element.nodes_id.end(); });
		if (current_iter != elements.end()) {
			answer.push_back(*current_iter);
			++current_iter;
		}
	}
	return answer;
}

std::set<Node> MeshLoader::Get_top_nodes_by_border_ID(int s_id) {
	std::set<Node> answer;
	for (auto a = boundary.begin(); a < boundary.end(); a++) {
		if (a->border_id == s_id) {
			for (auto b = a->nodes_id.begin(); b < a->nodes_id.end(); b++) {
				if (find_if(answer.begin(), answer.end(), [b](const Node &s) { //!!! Копирование s +
					return (s.id == *b);
					}) == answer.end()) {
					answer.insert(nodes.at((*b) - 1)); //!!! Где гарантия, что узлы не будут дублироваться - неверно выбран контейнер. +
				}
			}
		}
	}
	return answer;
}

std::vector<FiniteElement> MeshLoader::Get_Elements_by_ID_area(int a_id) const {
	std::vector<FiniteElement> answer; //!!! Как раз здесь дубликатов и так не будет. +
	auto current_iter = elements.begin();
	while (current_iter != elements.end()) {
		current_iter = std::find_if(current_iter, elements.end(), [a_id](const FiniteElement& s) { return s.material_id == a_id; });
		if (current_iter != elements.end()) {
			answer.push_back(*current_iter);
			++current_iter;
		}
	}
	return answer;
}

std::vector<BoundaryFiniteElement> MeshLoader::Get_Borders_by_Border_ID(int s_id) {
	std::vector<BoundaryFiniteElement> answer;
	auto surface_by_ID = boundary.begin();
	while (surface_by_ID != boundary.end()) {
		surface_by_ID = std::find_if(surface_by_ID, boundary.end(), [s_id](const BoundaryFiniteElement& s) {
			return s.border_id == s_id; });
		if (surface_by_ID != boundary.end()) {
			answer.push_back(*surface_by_ID);
			++surface_by_ID;
		}
	}
	return answer;
}

std::vector<std::set<int>>  MeshLoader::Get_Nodes_Neighbors() {
	std::vector<std::set<int>> result(nodes.size());
	for (const auto& FiniteElement : elements) {
		for (auto& nodes : FiniteElement.nodes_id) {
		    
			result[nodes - 1].insert(FiniteElement.nodes_id.begin(), FiniteElement.nodes_id.end());
			result[nodes - 1].erase(nodes);
		}
	}

	return result;
}

void MeshLoader::Insert_New_Nodes() {
	std::unordered_set<Edge, EdgeHash> edge_container;
	for (auto& elem : elements) {
		std::vector<int> current_iter(elem.nodes_id);
		for (auto i = current_iter.begin(); i != current_iter.end(); i++) {
			for (auto j = i + 1; j != current_iter.end(); j++) {
				Edge current_edge(*i, *j, 0);
				auto search_edge = edge_container.find(current_edge);

				if (search_edge == edge_container.end()) {
					Node left_node = Get_Node_by_ID(*i);
					Node right_node = Get_Node_by_ID(*j);
					Node center_node(nodes.size() + 1,
						std::vector<double>{(left_node.coord[0] + right_node.coord[0]) / 2,
						(left_node.coord[1] + right_node.coord[1]) / 2,
						(left_node.coord[2] + right_node.coord[2]) / 2},
						false);
					current_edge.center_id = nodes.size() + 1;
					edge_container.insert(current_edge);
					nodes.push_back(center_node);
					elem.nodes_id.push_back(current_edge.center_id);
				} else {
					elem.nodes_id.push_back(search_edge->center_id);
				}

			}
		}
	}

	for (auto& it : boundary) {
		std::vector<int> current_iter(it.nodes_id);
		for (auto i = current_iter.begin(); i != current_iter.end(); i++) {
			for (auto j = i + 1; j != current_iter.end(); j++) {
				Edge search_edge(*i, *j, 0);
				auto search_point = edge_container.find(search_edge);
				if (search_point != edge_container.end()) {
					it.nodes_id.push_back(search_point->center_id);
				}
			}
		}
	}
}

Node MeshLoader::Get_Node_by_ID(int n_id) {
	return nodes.at(n_id);
}